/* NiuTrans.Tensor - an open-source tensor library
* Copyright (C) 2017, Natural Language Processing Lab, Northestern University.
* All rights reserved.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*   http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

/*
* $Created by: Lin Ye (email: linye2015@outlook.com) 2018-06-14
*/

#include "TRectify.h"

namespace nts { // namespace nts(NiuTrans.Tensor)

/* 
case 1: test rectify function
In this case, y = max(0, x) 
*/
bool TestRectify1()
{
    /* a x tensor of size (2, 3) */
    int xOrder = 2;
    int * xDimSize = new int[xOrder];
    xDimSize[0] = 2;
    xDimSize[1] = 3;

    int xUnitNum = 1;
    for (int i = 0; i < xOrder; i++)
        xUnitNum *= xDimSize[i];

    /* a y tensor of size (2, 3) */
    int yOrder = 2;
    int * yDimSize = new int[yOrder];
    yDimSize[0] = 2;
    yDimSize[1] = 3;

    int yUnitNum = 1;
    for (int i = 0; i < yOrder; i++)
        yUnitNum *= yDimSize[i];

    DTYPE xData[2][3] = { {0.0F, -1.0F, 2.0F},
                          {3.0F, -4.0F, -5.0F} };
    DTYPE answer[2][3] = { {0.0F, 0.0F, 2.0F},
                           {3.0F, 0.0F, 0.0F} };

    /* CPU test */
    bool cpuTest = true;

    /* create tensors */
    XTensor * x = NewTensor(xOrder, xDimSize);
    XTensor * y = NewTensor(yOrder, yDimSize);

    /* initialize variables */
    x->SetData(xData, xUnitNum);
    y->SetZeroAll();

    /* call Rectify function */
    Rectify(x, y);

    /* check results */
    cpuTest = y->CheckData(answer, yUnitNum);

#ifdef USE_CUDA
	/* GPU test */
	bool gpuTest = true;

	/* create tensor */
	XTensor * xGPU = NewTensor(xOrder, xDimSize, X_FLOAT, 1.0F, 0);
	XTensor * yGPU = NewTensor(yOrder, yDimSize, X_FLOAT, 1.0F, 0);

	/* Initialize variables */
	xGPU->SetData(xData, xUnitNum);
	yGPU->SetZeroAll();

	/* call Rectify function */
	Rectify(xGPU, yGPU);

	/* check results */
	gpuTest = yGPU->CheckData(answer, yUnitNum);

	/* destroy variables */
	delete x;
    delete y;
    delete xGPU;
    delete yGPU;
	delete[] xDimSize;
    delete[] yDimSize;

	return cpuTest && gpuTest;
#else
	/* destroy variables */
	delete x;
    delete y;
	delete[] xDimSize;
    delete[] yDimSize;

	return cpuTest;
#endif // USE_CUDA
}

/* 
case 2: backward computation 
dE/dx = dE/dy * dy/dx 
rectified: y = max(0, x) 
In this case, lossName=CROSSENTROPY.
*/
bool TestRectify2()
{
	/* a x tensor of size (2, 3) */
	int xOrder = 2;
	int * xDimSize = new int[xOrder];
	xDimSize[0] = 2;
	xDimSize[1] = 3;

	int xUnitNum = 1;
	for (int i = 0; i < xOrder; i++)
		xUnitNum *= xDimSize[i];

	DTYPE xData[2][3] = { {1.0F, 1.0F, 2.0F},
	                      {2.0F, 4.0F, 5.0F} };
	DTYPE yData[2][3] = { {1.0F, 1.0F, 2.0F},
	                      {2.0F, 4.0F, 5.0F} };
	DTYPE goldData[2][3] = { {1.0F, 1.0F, 1.0F},
	                         {1.0F, 1.0F, 1.0F} };
	DTYPE dedyData[2][3] = { {-1.0F, -1.0F, -0.5F},
	                         {-0.5F, -0.25F, -0.2F} };
	DTYPE answer[2][3] = { {-1.0F, -1.0F, -0.5F},
	                       {-0.5F, -0.25F, -0.2F} };

	/* CPU test */
	bool cpuTest = true;

	/* create tensors */
	XTensor * x = NewTensor(xOrder, xDimSize);
	XTensor * y = NewTensor(xOrder, xDimSize);
	XTensor * gold = NewTensor(xOrder, xDimSize);
	XTensor * dedy = NewTensor(xOrder, xDimSize);
	XTensor * dedx = NewTensor(xOrder, xDimSize);

	/* initialize variables */
	x->SetData(xData, xUnitNum);
	y->SetData(yData, xUnitNum);
	gold->SetData(goldData, xUnitNum);
	dedy->SetData(dedyData, xUnitNum);
	dedx->SetZeroAll();

	/* call RectifyBackward function */
	RectifyBackward(gold, y, x, dedy, dedx, NOLOSS);

	/* check results */
	cpuTest = dedx->CheckData(answer, xUnitNum);

#ifdef USE_CUDA
	/* GPU test */
	bool gpuTest = true;

	/* create tensors */
	XTensor * xGPU = NewTensor(xOrder, xDimSize, X_FLOAT, 1.0F, 0);
	XTensor * yGPU = NewTensor(xOrder, xDimSize, X_FLOAT, 1.0F, 0);
	XTensor * goldGPU = NewTensor(xOrder, xDimSize, X_FLOAT, 1.0F, 0);
	XTensor * dedyGPU = NewTensor(xOrder, xDimSize, X_FLOAT, 1.0F, 0);
	XTensor * dedxGPU = NewTensor(xOrder, xDimSize, X_FLOAT, 1.0F, 0);

	/* initialize variables */
	xGPU->SetData(xData, xUnitNum);
	yGPU->SetData(yData, xUnitNum);
	goldGPU->SetData(goldData, xUnitNum);
	dedyGPU->SetData(dedyData, xUnitNum);
	dedxGPU->SetZeroAll();

	/* call rectifybackward function */
	RectifyBackward(goldGPU, yGPU, xGPU, dedyGPU, dedxGPU, NOLOSS);
    
	/* check results */
	gpuTest = dedxGPU->CheckData(answer, xUnitNum);

	/* destroy variables */
    delete x;
    delete y;
    delete dedy;
    delete dedx;
    delete gold;
    delete xGPU;
    delete yGPU;
    delete dedyGPU;
    delete dedxGPU;
    delete goldGPU;
	delete[] xDimSize;

	return cpuTest && gpuTest;
#else
	/* destroy variables */
    delete x;
    delete y;
    delete dedy;
    delete dedx;
    delete gold;
	delete[] xDimSize;

	return cpuTest;
#endif // USE_CUDA
}

/* other cases */
/*
TODO!!
*/

/* test for Rectify Function */
bool TestRectify()
{
    XPRINT(0, stdout, "[TEST RECTIFY] test rectify and its backward computation \n");
    bool returnFlag = true, caseFlag = true;

    /* case 1 test */
    caseFlag = TestRectify1();

    if (!caseFlag) {
        returnFlag = false;
        XPRINT(0, stdout, ">> case 1 failed!\n");
    }
    else
        XPRINT(0, stdout, ">> case 1 passed!\n");

    /* case 2 test */
    caseFlag = TestRectify2();

	if (!caseFlag) {
		returnFlag = false;
		XPRINT(0, stdout, ">> case 2 failed!\n");
	}
	else
		XPRINT(0, stdout, ">> case 2 passed!\n");

    /* other cases test */
    /*
    TODO!!
    */

    if (returnFlag) {
        XPRINT(0, stdout, ">> All Passed!\n");
    }
    else
        XPRINT(0, stdout, ">> Failed!\n");

    XPRINT(0, stdout, "\n");

    return returnFlag;
}

} // namespace nts(NiuTrans.Tensor)
